/*
 *  vectex - a SVG vector texture plugin for Blender
 *  Copyright (C) 2008 Marcelo de Gomensoro Malheiros <mgmalheiros@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>

#include "agg_pixfmt_rgb.h"
#include "agg_rasterizer_scanline_aa.h"
#include "agg_renderer_scanline.h"
#include "agg_rendering_buffer.h"
#include "agg_scanline_p.h"

#include "agg_svg_parser.h"
#include "vectex.h"
#include "MEM_guardedalloc.h"

/*---------------- Main declarations ----------------*/

#define MAX2(x,y)  ( (x)>(y) ?  (x) : (y) )

typedef struct Backend
{
    agg::svg::path_renderer path;
    double xmin;
    double ymin;
    double xmax;
    double ymax;
} Backend;


/*---------------- Main functions ----------------*/

EXT void *create_backend(char *filename)
{
    MSG0("[create_backend]\n");
    
    Backend *backend = new Backend();

    agg::svg::parser parser(backend->path);
    try
    {
        parser.parse(filename);
    }
    catch(agg::svg::exception& e)
    {
        MSG1("- backend error: %s\n", e.msg());
        delete backend;
        return (NULL);
    }
    backend->path.bounding_rect(&backend->xmin, &backend->ymin, &backend->xmax, &backend->ymax);

    return (backend);
}

EXT unsigned char *backend_render_tile(void *backend_data, int resolution,
                                       int tile_size, int tile_row, int tile_col,
                                       double base_r, double base_g, double base_b,
                                       int border)
{
    Backend *backend = (Backend *) backend_data;

    typedef agg::renderer_base<agg::pixfmt_rgb24>          renderer_base;
    typedef agg::renderer_scanline_aa_solid<renderer_base> renderer_solid;

    int width  = tile_size + border;
    int height = tile_size + border;
    
    /*
    unsigned char          *buf = (unsigned char *)memory_alloc(width * height * 3, "vectex:tile");
    */
    unsigned char          *buf = (unsigned char *)MEM_mallocN(width * height * 3, "vectex:tile");
    agg::rendering_buffer   rbuf(buf, width, height, width * 3);
    agg::pixfmt_rgb24       pixfmt(rbuf);
    renderer_base           rb(pixfmt);
    renderer_solid          ren(rb);

    rb.clear(agg::rgba(base_r, base_g, base_b));

    agg::rasterizer_scanline_aa<> ras;
    agg::scanline_p8              sl;
    agg::rect_i                   clip_box(0, 0, width, height);

    double size = MAX2(backend->xmax - backend->xmin, backend->ymax - backend->ymin); 
    agg::trans_affine mtx;
    mtx *= agg::trans_affine_translation((backend->xmin + backend->xmax) * -0.5,
                                         (backend->ymin + backend->ymax) * -0.5);
    mtx *= agg::trans_affine_scaling(resolution / size);
    mtx *= agg::trans_affine_translation(resolution * 0.5 - tile_size * tile_col,
                                         resolution * 0.5 - tile_size * tile_row);

    backend->path.expand(0.0);
    backend->path.render(ras, sl, ren, mtx, clip_box);
    
    return (buf); 
}

EXT void destroy_backend(void *backend_data)
{
    MSG0("[destroy_backend]\n");

    Backend *backend = (Backend *) backend_data;
    delete backend;
}
